@@ -1,6 +1,7 @@
 import datetime
 import logging
 import time
+from contextlib import contextmanager
 from operator import attrgetter
 
 from . import base
@@ -55,6 +56,18 @@ class NodeManager(base.NodeManagerBase):
         """Save an existing node"""
         return self.storage.save_node(node)
 
+    @contextmanager
+    def set_aggregating(self, node_id):
+        node = self.storage.set_aggregating(node_id)
+        start = datetime.datetime.utcnow()
+        try:
+            yield node
+        finally:
+            if node:
+                node.last_timeline_aggregation = start
+                node.is_aggregating = False
+                self.save_node(node)
+
 
 class ActivityManager(base.ActivityManagerBase):
     """Serializes :class:`Activity` objects."""
@@ -139,51 +152,56 @@ class Aggregator(object):
 
     def create_timeline(self, node_id):
         """Create, store, and return the timeline for a given node.
+
+        If an aggregation is already running for the given node, the most
+        recently aggregated timeline will be returned, and a new aggregation
+        will not be started.
+
         """
-        node = self.node_manager.get_node(node_id)
-        if not node:
-            log.warning('Node not found for node_id "%s". Creating...', node_id)
-            node = self.node_manager.create_node(node_id)
-        last_timeline_aggregation = node.last_timeline_aggregation
-        node.last_timeline_aggregation = datetime.datetime.utcnow()
-        # save node to persist last_timeline_aggregation timestamp
-        self.node_manager.save_node(node)
-        # get a subset of the nodes be followed
-        connections = self.filter_connections(
-                self.node_manager.get_nodes(node.following or []))
-        # retrieve the followed nodes' activities
-        activities = self.activity_manager.get_activities(connections,
-                since=last_timeline_aggregation)
-        # filter activities for followed nodes
-        activities = self.filter_activities(activities)
-        # add activities for this node
-        activities += self.activity_manager.get_activities([node],
-                since=last_timeline_aggregation)
-        # if we don't have any activities at this point, there's nothing from
-        # which to generate a timeline
-        if not activities:
-            return self.activity_manager.get_timeline(node_id)
-        # remove duplicates
-        activities = self._unique_activities(activities)
-        # classify and score activities
-        activities = self.classify_activities(activities)
-        activities = self.score_activities(activities)
-        # save to this node's timeline
-        self.activity_manager.save_timeline(node_id, activities)
-        return self.activity_manager.get_timeline(node_id)
+        with self.node_manager.set_aggregating(node_id) as node:
+            if node:
+                last_timeline_aggregation = node.last_timeline_aggregation
+
+                # get a subset of the nodes being followed
+                connections = []
+                if node.following:
+                    connections = self.filter_connections(
+                            self.node_manager.get_nodes(node.following))
+
+                activities = []
+                if connections:
+                    # retrieve the followed nodes' activities
+                    activities = self.activity_manager.get_activities(connections,
+                            since=last_timeline_aggregation)
+                    # filter activities for followed nodes
+                    activities = self.filter_activities(activities)
+
+                # add activities for this node
+                activities += self.activity_manager.get_activities([node],
+                        since=last_timeline_aggregation)
+                # if we don't have any activities at this point, there's nothing from
+                # which to generate a timeline
+                if not activities:
+                    return self.activity_manager.get_timeline(node_id)
+                # remove duplicates
+                activities = self._unique_activities(activities)
+                # classify and score activities
+                activities = self.classify_activities(activities)
+                activities = self.score_activities(activities)
+                # save to this node's timeline
+                self.activity_manager.save_timeline(node_id, activities)
 
     def needs_aggregation(self, node):
-        """Return True if it's time for this node's timeline to be
-        (re)aggregated).
-        """
-        # For now, just do an aggregation every time a timeline is requested
-        return True
+        """Return True if this node's timeline needs to be (re)aggregated).
 
-        # Alternative, based on time since last aggregation:
+        """
         last = node.last_timeline_aggregation
-        if not last: return True
-        # TODO: put this time threshhold in config
-        return (datetime.datetime.utcnow() - last).seconds > 60
+        nodes = [node]
+        if node.following:
+            nodes.extend(self.node_manager.get_nodes(node.following))
+        activities = self.activity_manager.get_activities(nodes, since=last,
+                limit=1)
+        return bool(activities)
 
     def get_timeline(self, node, page=0, limit=100, actor_only=False, filter_func=None):
         """Return a (paged and limited) timeline for `node`.
